var gWindow = {};


function loaded () {
  gWindow.browser = document.getElementById("browser");
}


function getController () {
  return gController;
}


var gController = {
  QueryInterface: function (iid) {
    if (iid.equals(Components.interfaces.nsISupports) ||
        iid.equals(Components.interfaces.nsISupportsWeakReference) ||
        iid.equals(Components.interfaces.nsIWebProgressListener))
      return this;
    throw Components.results.NS_NOINTERFACE;
  },
  supportsCommand: function (cmd) {
    switch (cmd) {
      case "cmd-download-project":
      case "cmd-upload-project":
        return true;
      default:
        return false;
    }
  },
  isCommandEnabled: function (cmd) {
    switch (cmd) {
      case "cmd-download-project":
      case "cmd-upload-project":
        return true;
      default:
        return false;
    }
  },
  doCommand: function (cmd) {
    switch (cmd) {
      case "cmd-download-project":
        top.goDoCommand(cmd);
        break;
      case "cmd-upload-project":
        cmdUploadProject();
        break;
    }
  },


  open: function (project, docres) {
    this.project = project;
    var uri = Cx.WEBPANEL_URL + project.id;
    if (docres && (docres instanceof Components.interfaces.nsIRDFResource)) {
      docres = docres.QueryInterface(Components.interfaces.nsIRDFResource);
      var rdfsvc = getRDFService();
      var srcarc = rdfsvc.GetResource(Cx.NS_DC + "source");
      var lit = project.localDS.GetTarget(docres, srcarc, true);
      if (lit && (lit instanceof Components.interfaces.nsIRDFLiteral)) {
        uri = lit.QueryInterface(Components.interfaces.nsIRDFLiteral).Value;
        project.localDS.Unassert(docres, srcarc, lit);
      }
    }

    gWindow.browser.addProgressListener(this);
    gWindow.browser.loadURI(uri, null, "UTF-8");
  },


  onLocationChange: function (prog, req, location) {},
  onProgressChange: function (prog, req, curself, maxself, curtot, maxtot) {},
  onSecurityChange: function (prog, req, state) {},
  onStateChange: function (prog, req, flags, status) {
    var IWeb = Components.interfaces.nsIWebProgressListener;
    // Different stop events come at different times, I've given up on
    // trying to figure out the right one to avoid incorrect Javascript
    // timing issues. Respond to 'em all.
    var targetFlags = IWeb.STATE_STOP;
    if ((flags & targetFlags) == targetFlags)
      this.setWindowProperties();
  },
  onStatusChange: function (prog, req, status, msg) {},


  setWindowProperties: function setWindowProperties () {
    gWindow.browser.contentWindow.cxws_clientVersion = Cx.VERSION;
    gWindow.browser.contentWindow.cxws_projectVersion = this.project.revision;
    gWindow.browser.contentWindow.cxws_projectTitle = this.project.title;
    gWindow.browser.contentWindow.cxws_projectURI = this.project.res.Value;
    gWindow.browser.contentWindow.cxws_projectModified =
      top.isProjectModified();

    gWindow.browser.contentWindow.startUpload = cmdUploadProject;
    gWindow.browser.contentWindow.resetProjectID = cmdResetProjectID;
    gWindow.browser.contentWindow.resetAllIDs = cmdResetAllIDs;
    gWindow.browser.contentWindow.createAccount = function () {
      var reg = top.gApp.getRegistration();
      if (! reg)
        return;
      var cxsvc = Components.classes["@celtx.com/celtx-service;1"]
        .getService(Components.interfaces.nsICeltxService);
      var observer = {
        onLogin: function (succeeded) {
          if (succeeded)
            top.openWebPanel();
        }
      };
      cxsvc.loginAs(reg.username, reg.password, "publish", observer, false);
    };
  },


  close: function close () {},
  save: function save () {},
  focus: function focus () {
    goUpdateCommand('cmd-download-project');
    goUpdateCommand('cmd-upload-project');
  },
  blur: function blur () {},


  setLocation: function setLocation (uri) {
    if (! uri)
      uri = Cx.WEBPANEL_URL + this.project.id;
    gWindow.browser.loadURI(uri, null, "UTF-8");
  }
};

function temporary () {
  return true;
}


function cmdResetProjectID () {
  gController.project.changeProjectID();
}


function cmdResetAllIDs () {
  gController.project.changeAllIDs();
}


function cmdUploadProject () {
  if (! gApp.authenticatedForService("publish-project")) {
    var callback = function (succeeded) {
      if (succeeded) goDoCommand("cmd-upload-project");
    };
    gApp.authenticate("publish-project", callback);
    return;
  }

  var project = gController.project;
  var publish = project.publishSettings;

  if (! publish.owner) publish.owner = gApp.userID;

  publish.canceled   = false;
  publish.authURL    = Cx.AUTH_URL;
  publish.publishURL = Cx.PUBLISH_URL;

  project.savePublishSettings(publish);
  top.gFrameLoader.saveAllFrames();
  project.modified = new Date();
  project.flush();

  var syncURL = Cx.SYNC_URL + (publish.force ? '/force' : '');

  var syncArg = {
    path:       project.projectFolder,
    comments:   publish.comments,
    syncURL:    syncURL,
    visibility: publish.mode,
    projectID:  project.id,
    projectURI: project.res.Value,
    title:      project.title,
    revision:   project.revision,
    modified:   top.isProjectModified,
    closeURL:   null
  };

  window.openDialog(Cx.CONTENT_PATH + 'sync.xul', '_blank',
                    Cx.MODAL_DIALOG_FLAGS, 'sync', syncArg);

  if (! syncArg.succeeded)
    return;

  project.setRevision(syncArg.revision);
  project.flush();
  project.isModified = true;
  if (syncArg.closeURL) {
    var IWeb = Components.interfaces.nsIWebNavigation;
    gWindow.browser.loadURIWithFlags(syncArg.closeURL,
      IWeb.LOAD_FLAGS_BYPASS_CACHE, null, "UTF-8");
  }
}
